package com.sleepycat.je.cleaner;

import java.util.Map;
import java.util.Set;

import com.sleepycat.je.DatabaseException;
import com.sleepycat.je.config.EnvironmentParams;
import com.sleepycat.je.dbi.DbConfigManager;
import com.sleepycat.je.dbi.DbTree;
import com.sleepycat.je.log.CleanerFileReader;
import com.sleepycat.je.tree.BIN;
import com.sleepycat.je.tree.LN;
import com.sleepycat.je.tree.TreeLocation;
import com.sleepycat.je.utilint.DbLsn;

public privileged aspect LookAheadCaching percflow(processFile(*)){

	pointcut processFile(FileProcessor fp):
		execution(boolean FileProcessor.processFile(Long)) && this(fp);

	LookAheadCache lookAheadCache = null;

	void around(Long fileNum, TreeLocation location, Long offset, LNInfo info,
			FileProcessor fp) throws DatabaseException:
		call(void FileProcessor.processLN(Long,TreeLocation,Long,LNInfo,Map)) && 
		args(fileNum, location,offset, info,Map) && this(fp) && 
		withincode(boolean hookr_processFileInternalLoop(Long , TrackedFileSummary , PackedOffsets.Iterator , long , Set , Map , CleanerFileReader , DbTree , TreeLocation )){

		if (lookAheadCache == null) {
			lookAheadCache = new LookAheadCache_Count(
					fp.cleaner.lookAheadCacheSize);
		}

		lookAheadCache.add(offset, info);

		if (lookAheadCache.isFull()) {
			Long poffset = lookAheadCache.nextOffset();
			LNInfo pinfo = lookAheadCache.remove(poffset);
			proceed(fileNum, location, poffset, pinfo, fp);

			LN ln = info.getLN();
			boolean isDupCountLN = ln.containsDuplicates();
			BIN bin = location.bin;
			int index = location.index;

			/*
			 * For all other non-deleted LNs in this BIN, lookup their LSN in
			 * the LN queue and process any matches.
			 */
			if (!isDupCountLN) {
				for (int i = 0; i < bin.getNEntries(); i += 1) {
					long lsn = bin.getLsn(i);
					if (i != index && !bin.isEntryKnownDeleted(i)
							&& !bin.isEntryPendingDeleted(i)
							&& DbLsn.getFileNumber(lsn) == fileNum.longValue()) {

						Long myOffset = new Long(DbLsn.getFileOffset(lsn));
						LNInfo myInfo = lookAheadCache.remove(myOffset);

						if (myInfo != null) {

							fp.processFoundLN(myInfo, lsn, lsn, bin, i, null);
						}
					}
				}
			}
		}
	}

	after(FileProcessor fp, Long fileNum, Map dbCache, TreeLocation location)
			throws DatabaseException:
		execution(boolean FileProcessor.hookr_processFileInternalLoop(Long , TrackedFileSummary , PackedOffsets.Iterator , long , Set , Map , CleanerFileReader , DbTree , TreeLocation )) && this(fp) &&
		args(fileNum,TrackedFileSummary , PackedOffsets.Iterator , long , Set ,dbCache,CleanerFileReader , DbTree,location){

		/* Process remaining queued LNs. */
		while (!lookAheadCache.isEmpty()) {
			fp.hook_beforeProcess();

			Long poffset = lookAheadCache.nextOffset();
			LNInfo pinfo = lookAheadCache.remove(poffset);
			fp.processLN(fileNum, location, poffset, pinfo, dbCache);
		}
	}

	int Cleaner.lookAheadCacheSize;

	after(Cleaner c, DbConfigManager cm) throws DatabaseException:
		execution(void Cleaner.envConfigUpdate(DbConfigManager)) && args(cm) && this(c){

		c.lookAheadCacheSize = cm
				.getInt(EnvironmentParams.CLEANER_LOOK_AHEAD_CACHE_SIZE);
	}

}
